@@ -18,41 +18,52 @@
  *   License along with this library.  If not, see <http://www.gnu.org/licenses/>.
  */
 #include "kwindowsystem.h"
-#include "kwindowsystem_p.h"
+#include "kwindowsystem_dummy_p.h"
+#include "kwindowsystemplugininterface_p.h"
+#include "pluginwrapper_p.h"
 
 #include <config-kwindowsystem.h>
 
-#if KWINDOWSYSTEM_HAVE_X11
-#include "kwindowsystem_p_x11.h"
-#endif
-#include "kwindowsystem_p_wayland.h"
-
 #include <QGuiApplication>
 #include <QMetaMethod>
 #include <QPixmap>
+#include <QPluginLoader>
 #include <QWidget>
 #include <QWindow>
+#if KWINDOWSYSTEM_HAVE_X11
+#include <QX11Info>
+#endif
+
+//QPoint and QSize all have handy / operators which are useful for scaling, positions and sizes for high DPI support
+//QRect does not, so we create one for internal purposes within this class
+inline QRect operator/(const QRect &rectangle, qreal factor)
+{
+    return QRect(rectangle.topLeft() / factor, rectangle.size() / factor);
+}
 
 class KWindowSystemStaticContainer
 {
 public:
     KWindowSystemStaticContainer() {
-#if KWINDOWSYSTEM_HAVE_X11
-        if (d.isNull() && (QGuiApplication::platformName() == QStringLiteral("xcb"))) {
-            d.reset(new KWindowSystemPrivateX11());
-        }
-#endif
-        if (d.isNull() && (QGuiApplication::platformName().startsWith(QLatin1String("wayland")))) {
-            d.reset(new KWindowSystemPrivateWayland());
-        }
-        if (d.isNull()) {
-            d.reset(new KWindowSystemPrivateDummy());
-        }
+        d.reset(KWindowSystemPluginWrapper::self().createWindowSystem());
 
-        kwm.moveToThread(QCoreApplication::instance()->thread());
+        if (QCoreApplication::instance()) {
+            kwm.moveToThread(QCoreApplication::instance()->thread());
+        }
+    }
+    KWindowSystemPrivate *xcbPlugin() {
+        if (xcbPrivate.isNull()) {
+            QPluginLoader loader(QStringLiteral(XCB_PLUGIN_PATH));
+            QScopedPointer<KWindowSystemPluginInterface> xcbPlugin(qobject_cast< KWindowSystemPluginInterface* >(loader.instance()));
+            if (!xcbPlugin.isNull()) {
+                xcbPrivate.reset(xcbPlugin->createWindowSystem());
+            }
+        }
+        return xcbPrivate.data();
     }
     KWindowSystem kwm;
     QScopedPointer<KWindowSystemPrivate> d;
+    QScopedPointer<KWindowSystemPrivate> xcbPrivate;
 };
 
 Q_GLOBAL_STATIC(KWindowSystemStaticContainer, g_kwmInstanceContainer)
@@ -61,6 +72,16 @@ KWindowSystemPrivate::~KWindowSystemPrivate()
 {
 }
 
+QPixmap KWindowSystemPrivate::iconFromNetWinInfo(int width, int height, bool scale, int flags, NETWinInfo *info)
+{
+    Q_UNUSED(width)
+    Q_UNUSED(height)
+    Q_UNUSED(scale)
+    Q_UNUSED(flags)
+    Q_UNUSED(info)
+    return QPixmap();
+}
+
 QList<WId> KWindowSystemPrivateDummy::windows()
 {
     return QList<WId>();
@@ -451,6 +472,10 @@ void KWindowSystem::setMainWindow(QWidget *subWidget, WId mainWindowId)
     Q_ASSERT(subWindow);
 
     QWindow *mainWindow = QWindow::fromWinId(mainWindowId);
+    if (!mainWindow) {
+        // foreign windows not supported on all platforms
+        return;
+    }
     // mainWindow is not the child of any object, so make sure it gets deleted at some point
     connect(subWidget, &QObject::destroyed, mainWindow, &QObject::deleteLater);
     subWindow->setTransientParent(mainWindow);
@@ -478,9 +503,19 @@ QPixmap KWindowSystem::icon(WId win, int width, int height, bool scale, int flag
 QPixmap KWindowSystem::icon(WId win, int width, int height, bool scale, int flags, NETWinInfo *info)
 {
     Q_D(KWindowSystem);
+    width *= qApp->devicePixelRatio();
+    height *= qApp->devicePixelRatio();
 #if KWINDOWSYSTEM_HAVE_X11
     if (info) {
-        return KWindowSystemPrivateX11::icon(width, height, scale, flags, info);
+        if (isPlatformX11()) {
+            // this is the xcb plugin, we can just delegate
+            return d->iconFromNetWinInfo(width, height, scale, flags, info);
+        } else {
+            // other platform plugin, load xcb plugin to delegate to it
+            if (KWindowSystemPrivate *p = g_kwmInstanceContainer()->xcbPlugin()) {
+                return p->iconFromNetWinInfo(width, height, scale, flags, info);
+            }
+        }
     }
 #else
     Q_UNUSED(info)
@@ -561,13 +596,13 @@ bool KWindowSystem::compositingActive()
 QRect KWindowSystem::workArea(int desktop)
 {
     Q_D(KWindowSystem);
-    return d->workArea(desktop);
+    return d->workArea(desktop) / qApp->devicePixelRatio();
 }
 
 QRect KWindowSystem::workArea(const QList<WId> &exclude, int desktop)
 {
     Q_D(KWindowSystem);
-    return d->workArea(exclude, desktop);
+    return d->workArea(exclude, desktop) / qApp->devicePixelRatio();
 }
 
 QString KWindowSystem::desktopName(int desktop)
@@ -605,15 +640,17 @@ void KWindowSystem::setExtendedStrut(WId win, int left_width, int left_start, in
                                      int bottom_width, int bottom_start, int bottom_end)
 {
     Q_D(KWindowSystem);
-    d->setExtendedStrut(win, left_width, left_start, left_end,
-                        right_width, right_start, right_end, top_width, top_start, top_end,
-                        bottom_width, bottom_start, bottom_end);
+    const qreal dpr = qApp->devicePixelRatio();
+    d->setExtendedStrut(win, left_width * dpr, left_start * dpr, left_end * dpr,
+                        right_width * dpr, right_start * dpr, right_end * dpr, top_width * dpr, top_start * dpr, top_end * dpr,
+                        bottom_width * dpr, bottom_start * dpr, bottom_end * dpr);
 }
 
 void KWindowSystem::setStrut(WId win, int left, int right, int top, int bottom)
 {
     Q_D(KWindowSystem);
-    d->setStrut(win, left, right, top, bottom);
+    const qreal dpr = qApp->devicePixelRatio();
+    d->setStrut(win, left * dpr, right * dpr, top * dpr, bottom * dpr);
 }
 
 bool KWindowSystem::icccmCompliantMappingState()
@@ -655,13 +692,13 @@ bool KWindowSystem::mapViewport()
 int KWindowSystem::viewportToDesktop(const QPoint &p)
 {
     Q_D(KWindowSystem);
-    return d->viewportToDesktop(p);
+    return d->viewportToDesktop(p / qApp->devicePixelRatio());
 }
 
 int KWindowSystem::viewportWindowToDesktop(const QRect &r)
 {
     Q_D(KWindowSystem);
-    return d->viewportWindowToDesktop(r);
+    return d->viewportWindowToDesktop(r / qApp->devicePixelRatio());
 }
 
 QPoint KWindowSystem::desktopToViewport(int desktop, bool absolute)
@@ -673,5 +710,42 @@ QPoint KWindowSystem::desktopToViewport(int desktop, bool absolute)
 QPoint KWindowSystem::constrainViewportRelativePosition(const QPoint &pos)
 {
     Q_D(KWindowSystem);
-    return d->constrainViewportRelativePosition(pos);
+    return d->constrainViewportRelativePosition(pos / qApp->devicePixelRatio());
+}
+
+static inline KWindowSystem::Platform initPlatform()
+{
+    auto platformName = QGuiApplication::platformName();
+    if (platformName == QLatin1String("flatpak")) {
+        // here we cannot know what is the actual windowing system, let's try it's env variable
+        const auto flatpakPlatform = QString::fromLocal8Bit(qgetenv("QT_QPA_FLATPAK_PLATFORM"));
+        if (!flatpakPlatform.isEmpty()) {
+            platformName = flatpakPlatform;
+        }
+    }
+#if KWINDOWSYSTEM_HAVE_X11
+    if (platformName == QLatin1String("xcb")) {
+        return KWindowSystem::Platform::X11;
+    }
+#endif
+    if (platformName.startsWith(QLatin1String("wayland"), Qt::CaseInsensitive)) {
+        return KWindowSystem::Platform::Wayland;
+    }
+    return KWindowSystem::Platform::Unknown;
+}
+
+KWindowSystem::Platform KWindowSystem::platform()
+{
+    static Platform s_platform = initPlatform();
+    return s_platform;
+}
+
+bool KWindowSystem::isPlatformX11()
+{
+    return platform() == Platform::X11;
+}
+
+bool KWindowSystem::isPlatformWayland()
+{
+    return platform() == Platform::Wayland;
 }
